/****************************************************************************
** Meta object code from reading C++ file 'qhttpserverresponder.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qthttpserver/src/httpserver/qhttpserverresponder.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qhttpserverresponder.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN20QHttpServerResponderE_t {};
} // unnamed namespace

template <> constexpr inline auto QHttpServerResponder::qt_create_metaobjectdata<qt_meta_tag_ZN20QHttpServerResponderE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QHttpServerResponder",
        "StatusCode",
        "Continue",
        "SwitchingProtocols",
        "Processing",
        "Ok",
        "Created",
        "Accepted",
        "NonAuthoritativeInformation",
        "NoContent",
        "ResetContent",
        "PartialContent",
        "MultiStatus",
        "AlreadyReported",
        "IMUsed",
        "MultipleChoices",
        "MovedPermanently",
        "Found",
        "SeeOther",
        "NotModified",
        "UseProxy",
        "TemporaryRedirect",
        "PermanentRedirect",
        "BadRequest",
        "Unauthorized",
        "PaymentRequired",
        "Forbidden",
        "NotFound",
        "MethodNotAllowed",
        "NotAcceptable",
        "ProxyAuthenticationRequired",
        "RequestTimeout",
        "Conflict",
        "Gone",
        "LengthRequired",
        "PreconditionFailed",
        "PayloadTooLarge",
        "UriTooLong",
        "UnsupportedMediaType",
        "RequestRangeNotSatisfiable",
        "ExpectationFailed",
        "ImATeapot",
        "MisdirectedRequest",
        "UnprocessableEntity",
        "Locked",
        "FailedDependency",
        "UpgradeRequired",
        "PreconditionRequired",
        "TooManyRequests",
        "RequestHeaderFieldsTooLarge",
        "UnavailableForLegalReasons",
        "InternalServerError",
        "NotImplemented",
        "BadGateway",
        "ServiceUnavailable",
        "GatewayTimeout",
        "HttpVersionNotSupported",
        "VariantAlsoNegotiates",
        "InsufficientStorage",
        "LoopDetected",
        "NotExtended",
        "NetworkAuthenticationRequired",
        "NetworkConnectTimeoutError"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'StatusCode'
        QtMocHelpers::EnumData<enum StatusCode>(1, 1, QMC::EnumIsScoped).add({
            {    2, StatusCode::Continue },
            {    3, StatusCode::SwitchingProtocols },
            {    4, StatusCode::Processing },
            {    5, StatusCode::Ok },
            {    6, StatusCode::Created },
            {    7, StatusCode::Accepted },
            {    8, StatusCode::NonAuthoritativeInformation },
            {    9, StatusCode::NoContent },
            {   10, StatusCode::ResetContent },
            {   11, StatusCode::PartialContent },
            {   12, StatusCode::MultiStatus },
            {   13, StatusCode::AlreadyReported },
            {   14, StatusCode::IMUsed },
            {   15, StatusCode::MultipleChoices },
            {   16, StatusCode::MovedPermanently },
            {   17, StatusCode::Found },
            {   18, StatusCode::SeeOther },
            {   19, StatusCode::NotModified },
            {   20, StatusCode::UseProxy },
            {   21, StatusCode::TemporaryRedirect },
            {   22, StatusCode::PermanentRedirect },
            {   23, StatusCode::BadRequest },
            {   24, StatusCode::Unauthorized },
            {   25, StatusCode::PaymentRequired },
            {   26, StatusCode::Forbidden },
            {   27, StatusCode::NotFound },
            {   28, StatusCode::MethodNotAllowed },
            {   29, StatusCode::NotAcceptable },
            {   30, StatusCode::ProxyAuthenticationRequired },
            {   31, StatusCode::RequestTimeout },
            {   32, StatusCode::Conflict },
            {   33, StatusCode::Gone },
            {   34, StatusCode::LengthRequired },
            {   35, StatusCode::PreconditionFailed },
            {   36, StatusCode::PayloadTooLarge },
            {   37, StatusCode::UriTooLong },
            {   38, StatusCode::UnsupportedMediaType },
            {   39, StatusCode::RequestRangeNotSatisfiable },
            {   40, StatusCode::ExpectationFailed },
            {   41, StatusCode::ImATeapot },
            {   42, StatusCode::MisdirectedRequest },
            {   43, StatusCode::UnprocessableEntity },
            {   44, StatusCode::Locked },
            {   45, StatusCode::FailedDependency },
            {   46, StatusCode::UpgradeRequired },
            {   47, StatusCode::PreconditionRequired },
            {   48, StatusCode::TooManyRequests },
            {   49, StatusCode::RequestHeaderFieldsTooLarge },
            {   50, StatusCode::UnavailableForLegalReasons },
            {   51, StatusCode::InternalServerError },
            {   52, StatusCode::NotImplemented },
            {   53, StatusCode::BadGateway },
            {   54, StatusCode::ServiceUnavailable },
            {   55, StatusCode::GatewayTimeout },
            {   56, StatusCode::HttpVersionNotSupported },
            {   57, StatusCode::VariantAlsoNegotiates },
            {   58, StatusCode::InsufficientStorage },
            {   59, StatusCode::LoopDetected },
            {   60, StatusCode::NotExtended },
            {   61, StatusCode::NetworkAuthenticationRequired },
            {   62, StatusCode::NetworkConnectTimeoutError },
        }),
    };
    return QtMocHelpers::metaObjectData<QHttpServerResponder, qt_meta_tag_ZN20QHttpServerResponderE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QHttpServerResponder::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QHttpServerResponderE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QHttpServerResponderE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN20QHttpServerResponderE_t>.metaTypes,
    nullptr
} };

QT_WARNING_POP
