/****************************************************************************
** Meta object code from reading C++ file 'qhttpserverrequest.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qthttpserver/src/httpserver/qhttpserverrequest.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qhttpserverrequest.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN18QHttpServerRequestE_t {};
} // unnamed namespace

template <> constexpr inline auto QHttpServerRequest::qt_create_metaobjectdata<qt_meta_tag_ZN18QHttpServerRequestE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QHttpServerRequest",
        "Method",
        "Unknown",
        "Get",
        "Put",
        "Delete",
        "Post",
        "Head",
        "Options",
        "Patch",
        "Connect",
        "Trace",
        "AnyKnown",
        "Methods"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'Method'
        QtMocHelpers::EnumData<enum Method>(1, 1, QMC::EnumIsScoped).add({
            {    2, Method::Unknown },
            {    3, Method::Get },
            {    4, Method::Put },
            {    5, Method::Delete },
            {    6, Method::Post },
            {    7, Method::Head },
            {    8, Method::Options },
            {    9, Method::Patch },
            {   10, Method::Connect },
            {   11, Method::Trace },
            {   12, Method::AnyKnown },
        }),
        // flag 'Methods'
        QtMocHelpers::EnumData<Methods>(13, 1, QMC::EnumIsFlag | QMC::EnumIsScoped).add({
            {    2, Method::Unknown },
            {    3, Method::Get },
            {    4, Method::Put },
            {    5, Method::Delete },
            {    6, Method::Post },
            {    7, Method::Head },
            {    8, Method::Options },
            {    9, Method::Patch },
            {   10, Method::Connect },
            {   11, Method::Trace },
            {   12, Method::AnyKnown },
        }),
    };
    return QtMocHelpers::metaObjectData<QHttpServerRequest, qt_meta_tag_ZN18QHttpServerRequestE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QHttpServerRequest::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN18QHttpServerRequestE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN18QHttpServerRequestE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN18QHttpServerRequestE_t>.metaTypes,
    nullptr
} };

QT_WARNING_POP
